% CSC320S 2005: Assignment 2
%
% Student: Jason Montojo
% Username: g2cookie@cdf

function ims = morphImagesLoop(im0, im1, linePairs, s)
% ims = morphImagesLoop(im0, im1, linePairs, s)
%
% Returns the image of the morph from im0 into im1 at time s,
% as defined by the control lines in linePairs.  The morph is computed by
% looping over each pixel location in the destination image.
%
%   s must be in [0, 1].
%
%   im0 and im1 must have the same dimensions.
%
%   linePairs is an 8xj matrix where j is the number of line segment pairs.
%   Each column of linePairs is of the form:
%
%     [Px0 Py0 Qx0 Qy0 Px1 Py1 Qx1 Qy1]'
%
%   ... where P0Q0 is a segment in im0 which corresponds to
%   P1Q1, a segment in im1.

a = eps;
b = 2;
p = 0.5;

ims = zeros(size(im0));
[W0, W1] = morphWarps(linePairs, size(im0), s, a, b, p);

for j = 1:size(im0, 1),
    for i = 1:size(im0, 2),
        % Determine the color from image 0 based on the warp function.
        c0 = bicubic(im0, squeeze(W0(j, i, :)));
        
        % Determine the color from image 1 based on the warp function.
        c1 = bicubic(im1, squeeze(W1(j, i, :)));
        
        % Cross-dissolve between the two colours.
        ims(j, i) = (1 - s) * c0 + s * c1;
    end
end

return
      
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function z = bicubic(data, point)
% z = bicubic(data, point)
%
% Determines the value at the given point within the matrix, data, through
% bicubic interpolation.  

[szY, szX] = size(data);

% Determine the discrete pixel location to the upper-left of the
% desired point.
origin = fix(point);

xTaps = (origin(1) + [-1 0 1 2]);
yTaps = (origin(2) + [-1 0 1 2]);

% Have the pixels at the border of the image repeat their value towards the
% outside of the image.
xTaps(find(xTaps < 1)) = 1;
xTaps(find(xTaps > szX)) = szX;
yTaps(find(yTaps < 1)) = 1;
yTaps(find(yTaps > szY)) = szY;
 
% x contains the 4x4 neighbourhood of pixels for that will be
% considered during the interpolation.
x = data(yTaps, xTaps);

t = point - origin;

% Perform 4 interpolations along the X dimension, in parallel.
xi = x * catmullRom(t(1))';

% Use the values from the previous 4 interpolations to perform the final
% interpolation along the Y dimension.
z = xi' * catmullRom(t(2))';
return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function filter = catmullRom(t)
% filter = catmullRom(t)
%
% Computes the filter matrix for a 4-tap Catmull-Rom interpolation at time t,
% where t in [0, 1].

coefficients = [0 0 2 0; 0 1 0 -1; -1 4 -5 2; 1 -3 3 -1];
filter = 0.5 * [1 1-t (1-t)^2 (1-t)^3] * coefficients;

return
