% CSC320S 2005: Assignment 2
%
% Student: Jason Montojo
% Username: g2cookie@cdf

function [W0, W1] = morphWarps(linePairs, szIm, s, a, b, p)
% [W0, W1] = morphWarps(linePairs, szIm, s, a, b, p)
% 
% Computes the warp functions W_{k,s} for k = 0, 1.
%
%   W0(j,i,d) gives the component of the d-th dimension of pixel location (i,j)
%     In other words, if pixel x=[i j]' in source image 0, then the warped
%     pixel z=[W0(j, i, 0) W0(j, i, 1)]'
%
%   linePairs is an 8 x n matrix of the n pairs of line segments that define
%     how an image should be warped.
%     Each column of linePairs is of the form:
%
%       [Px0 Py0 Qx0 Qy0 Px1 Py1 Qx1 Qy1]'
%
%     ... where P0Q0 is a segment in im0 which corresponds to
%     P1Q1, a segment in im1.
%
%   szIm is a vector containing the dimensions of the source image.
%
%   a, b, and p are parameters which affect the behaviour of the warp during
%     the weighting process.  For good results, the following constraints
%     should be observed:  a in (0,1], b in [0.5,2], p in [0,1].

% Compute pixel locations as a 2 x prod(szIm) array,
% where row 1 of x contains X components,
% and row 2 contains Y components
[p1, p2] = meshgrid(1:szIm(2), 1:szIm(1));
x = [p1(:)'; p2(:)'];

W0Sum = zeros([2 size(x, 2)]);
W1Sum = W0Sum;

% Make a signal extender to facilitate array operations.
xExtender = ones([1 size(x, 2)]);

mSum = zeros([1 size(x, 2)]);

for j = 1:size(linePairs, 2),
    P0 = linePairs([1 2], j);
    Q0 = linePairs([3 4], j);
    P1 = linePairs([5 6], j);
    Q1 = linePairs([7 8], j);

    % P_{s,j}
    P = (1 - s) * P0 + s * P1;
    
    % Q_{s,j}
    Q = (1 - s) * Q0 + s * Q1;
    
    % L_{s,j}
    L = norm(Q - P, 2);

    % t_{s,j}
    t = (Q - P) / L;
    
    % n_{s,j}
    n = [0 -1; 1 0] * t;
    
    xMinusP = x - P * xExtender;
    u = t' * xMinusP ./ L;
    v = n' * xMinusP;
    
    d = zeros([1 size(x, 2)]);
    d = d + (u > 1) .* sqrt(sum((x - Q * xExtender) .^ 2));
    d = d + (u >= 0 & u <= 1) .* abs(v);
    d = d + (u < 0) .* sqrt(sum((x - P * xExtender) .^ 2));
    
    % Compute n_{k,j} for k = 0, 1
    n0 = [0 -1; 1 0] * ((Q0 - P0) / norm(Q0 - P0, 2));
    n1 = [0 -1; 1 0] * ((Q1 - P1) / norm(Q1 - P1, 2));
    
    % Accumulate W_{k,s,j} for k = 0, 1
    W0j = P0 * xExtender + (Q0 - P0) * u + n0 * v;
    W1j = P1 * xExtender + (Q1 - P1) * u + n1 * v;

    % m_{s,j}
    mj = (L ^ p ./ (a + d)) .^ b;
    
    mSum = mSum + mj;
    
    mjExtended = [1; 1] * mj;
    W0Sum = W0Sum + mjExtended .* W0j;
    W1Sum = W1Sum + mjExtended .* W1j;
end

% Extend m to span 2 rows so we can do array operations
mExtended = [1; 1] * mSum;
W0Sum = W0Sum ./ mExtended;
W1Sum = W1Sum ./ mExtended;

% Pack the X-components into layer 1 of W0
W0(:, :, 1) = reshape(W0Sum(1, :), szIm);
W1(:, :, 1) = reshape(W1Sum(1, :), szIm);

% Pack the Y-components into layer 2 of W1
W0(:, :, 2) = reshape(W0Sum(2, :), szIm);
W1(:, :, 2) = reshape(W1Sum(2, :), szIm);
